import traceback
from enum import Enum
from typing import Optional, List

# noinspection PyProtectedMember
import sys
from pyspark.sql.types import (
    StructType,
    StructField,
    ArrayType,
    DataType,
    NullType,
    StringType,
    BinaryType,
    IntegralType,
    BooleanType,
    DateType,
    TimestampType,
    NumericType,
)

from spark_data_frame_comparer.schema_comparison_item import SchemaComparisonItem
from spark_data_frame_comparer.sparjk_data_frame_comparer_generic_exception import (
    SparkDataFrameComparerGenericException,
)


class SchemaCompareErrorType(Enum):
    INFO = 1
    WARNING = 2
    ERROR = 3
    FATAL = 4


class SchemaCompareError:
    def __init__(
        self,
        column: Optional[str],
        error_type: SchemaCompareErrorType,
        error: str,
        source_schema: DataType,
        desired_schema: DataType,
    ) -> None:
        self.column: Optional[str] = column
        self.error_type: SchemaCompareErrorType = error_type
        self.error: str = error
        self.source_schema: DataType = source_schema
        self.desired_schema: DataType = desired_schema

    def __str__(self) -> str:
        return (
            (("column: [" + self.column + "]") if self.column else "")
            + "\n"
            + str(self.error_type)
            + ": "
            + self.error
            + "\n"
            + "source_schema: "
            + str(self.source_schema)
            + "\n"
            + "desired_schema: "
            + str(self.desired_schema)
            + "\n"
        )


class SchemaComparerResult:
    def __init__(self) -> None:
        self.errors: List[SchemaCompareError] = []

    def __str__(self) -> str:
        return "\n".join([str(e) for e in self.errors])

    def to_string(self, include_info: bool = False) -> str:
        if include_info is True:
            return "\n".join([str(e) for e in self.errors])
        else:
            return "\n".join(
                [
                    str(e)
                    for e in self.errors
                    if e.error_type != SchemaCompareErrorType.INFO
                ]
            )


class SchemaComparer:
    @staticmethod
    def compare_struct(
        parent_column_name: Optional[str],
        source_schema: DataType,
        desired_schema: StructType,
        schema_comparison: SchemaComparisonItem,
        allow_missing_nullable_properties: bool,
    ) -> List[SchemaCompareError]:
        if isinstance(source_schema, NullType):
            return []
        if not isinstance(source_schema, StructType):
            return [
                SchemaCompareError(
                    column=parent_column_name,
                    error_type=SchemaCompareErrorType.ERROR,
                    error=f"Type of {parent_column_name} does not match with struct.",
                    source_schema=source_schema,
                    desired_schema=desired_schema,
                )
            ]

        errors: List[SchemaCompareError] = []
        desired_field: StructField
        i: int = 0
        for desired_field in desired_schema.fields:
            # first see if the field exists in source_schema.
            # if destination field is nullable then it is fine if there is no source field
            if i >= len(source_schema.names):
                # nothing to match so skip
                if not desired_field.nullable or not allow_missing_nullable_properties:
                    errors.append(
                        SchemaCompareError(
                            column=f"{parent_column_name}.{desired_field.name}",
                            error_type=SchemaCompareErrorType.ERROR,
                            error=f"{parent_column_name}.{desired_field.name} not found in source"
                            + (
                                " and is not nullable"
                                if allow_missing_nullable_properties
                                else " and is within an array so even nullable properties must be present"
                            ),
                            source_schema=NullType(),
                            desired_schema=desired_field.dataType,
                        )
                    )
            elif desired_field.name != source_schema.names[i]:
                if desired_field.name in source_schema.names:
                    # it's in wrong place
                    errors.append(
                        SchemaCompareError(
                            column=f"{parent_column_name}.{desired_field.name}",
                            error_type=SchemaCompareErrorType.ERROR,
                            error=f"{parent_column_name}.{desired_field.name} is not in the same order as source."
                            f" source=[{','.join(source_schema.names)}]"
                            f" destination=[{','.join(desired_schema.names)}]",
                            source_schema=NullType(),
                            desired_schema=desired_field.dataType,
                        )
                    )
                # if field is nullable then it's ok
                elif desired_field.nullable and allow_missing_nullable_properties:
                    errors.append(
                        SchemaCompareError(
                            column=f"{parent_column_name}.{desired_field.name}",
                            error_type=SchemaCompareErrorType.INFO,
                            error=f"{parent_column_name}.{desired_field.name} not found in source"
                            + (
                                " and is not nullable"
                                if allow_missing_nullable_properties
                                else " and is within an array so even nullable properties must be present"
                            ),
                            source_schema=NullType(),
                            desired_schema=desired_field.dataType,
                        )
                    )
                else:
                    errors.append(
                        SchemaCompareError(
                            column=f"{parent_column_name}.{desired_field.name}",
                            error_type=SchemaCompareErrorType.ERROR,
                            error=f"{parent_column_name}.{desired_field.name} not found in source and is not nullable",
                            source_schema=NullType(),
                            desired_schema=desired_field.dataType,
                        )
                    )
            else:
                i += 1  # match found so increment source
                source_field: StructField = source_schema[desired_field.name]
                errors += SchemaComparer.compare_data_type(
                    parent_column_name=f"{parent_column_name}.{desired_field.name}",
                    source_schema=source_field.dataType,
                    desired_schema=desired_field.dataType,
                    schema_comparison=schema_comparison,
                    allow_missing_nullable_properties=allow_missing_nullable_properties,
                )

        # now check if source has extra columns
        for source_field in source_schema.fields:
            if source_field.name not in [
                desired_field.name for desired_field in desired_schema.fields
            ]:
                errors.append(
                    SchemaCompareError(
                        column=f"{parent_column_name}.{source_field.name}",
                        error_type=SchemaCompareErrorType.ERROR,
                        error=f"{parent_column_name}.{source_field.name} is in source schema"
                        " but not in destination schema",
                        source_schema=source_field.dataType,
                        desired_schema=NullType(),
                    )
                )

        return errors

    @staticmethod
    def compare_array(
        parent_column_name: Optional[str],
        source_schema: DataType,
        desired_schema: ArrayType,
        schema_comparison: SchemaComparisonItem,
        allow_missing_nullable_properties: bool,
    ) -> List[SchemaCompareError]:
        if isinstance(source_schema, NullType):
            return []

        if not isinstance(source_schema, ArrayType):
            return [
                SchemaCompareError(
                    column=parent_column_name,
                    error_type=SchemaCompareErrorType.ERROR,
                    error=f"Type of {parent_column_name} does not match with array.",
                    source_schema=source_schema,
                    desired_schema=desired_schema,
                )
            ]

        return SchemaComparer.compare_data_type(
            parent_column_name=parent_column_name,
            source_schema=source_schema.elementType,
            desired_schema=desired_schema.elementType,
            schema_comparison=schema_comparison,
            allow_missing_nullable_properties=False,  # properties inside an array have to match exactly
        )

    @staticmethod
    def compare_simple(
        parent_column_name: Optional[str],
        source_schema: DataType,
        desired_schema: DataType,
        schema_comparison: SchemaComparisonItem,
        allow_missing_nullable_properties: bool,
    ) -> List[SchemaCompareError]:
        if isinstance(source_schema, NullType):
            return []

        # compare the two
        if desired_schema != source_schema:
            if SchemaComparer.can_cast(
                source_type=source_schema, desired_type=desired_schema
            ):
                return [
                    SchemaCompareError(
                        column=parent_column_name,
                        error_type=SchemaCompareErrorType.INFO,
                        error=f"Type of {parent_column_name} does not match with simple type but can be casted.",
                        source_schema=source_schema,
                        desired_schema=desired_schema,
                    )
                ]
            else:
                return [
                    SchemaCompareError(
                        column=parent_column_name,
                        error_type=SchemaCompareErrorType.ERROR,
                        error=f"Type of {parent_column_name} does not match with simple type.",
                        source_schema=source_schema,
                        desired_schema=desired_schema,
                    )
                ]
        return []

    @staticmethod
    def compare_data_type(
        parent_column_name: Optional[str],
        source_schema: DataType,
        desired_schema: DataType,
        schema_comparison: SchemaComparisonItem,
        allow_missing_nullable_properties: bool,
    ) -> List[SchemaCompareError]:
        if isinstance(desired_schema, StructType):
            return SchemaComparer.compare_struct(
                parent_column_name=parent_column_name,
                source_schema=source_schema,
                desired_schema=desired_schema,
                schema_comparison=schema_comparison,
                allow_missing_nullable_properties=allow_missing_nullable_properties,
            )
        elif isinstance(desired_schema, ArrayType):
            return SchemaComparer.compare_array(
                parent_column_name=parent_column_name,
                source_schema=source_schema,
                desired_schema=desired_schema,
                schema_comparison=schema_comparison,
                allow_missing_nullable_properties=allow_missing_nullable_properties,
            )
        else:
            return SchemaComparer.compare_simple(
                parent_column_name=parent_column_name,
                source_schema=source_schema,
                desired_schema=desired_schema,
                schema_comparison=schema_comparison,
                allow_missing_nullable_properties=allow_missing_nullable_properties,
            )

    @staticmethod
    def can_cast(source_type: DataType, desired_type: DataType) -> bool:
        if isinstance(source_type, NullType):
            return True

        if isinstance(desired_type, StringType):
            return True

        if isinstance(source_type, StringType) and isinstance(desired_type, BinaryType):
            return True

        if isinstance(source_type, IntegralType) and isinstance(
            desired_type, BinaryType
        ):
            return True

        if isinstance(source_type, StringType) and isinstance(
            desired_type, BooleanType
        ):
            return True

        if isinstance(source_type, DateType) and isinstance(desired_type, BooleanType):
            return True

        if isinstance(source_type, TimestampType) and isinstance(
            desired_type, BooleanType
        ):
            return True

        if isinstance(source_type, NumericType) and isinstance(
            desired_type, BooleanType
        ):
            return True

        if isinstance(source_type, StringType) and isinstance(
            desired_type, TimestampType
        ):
            return True

        if isinstance(source_type, BooleanType) and isinstance(
            desired_type, TimestampType
        ):
            return True

        if isinstance(source_type, DateType) and isinstance(
            desired_type, TimestampType
        ):
            return True

        if isinstance(source_type, NumericType) and isinstance(
            desired_type, TimestampType
        ):
            return True

        if isinstance(source_type, StringType) and isinstance(desired_type, DateType):
            return True

        if isinstance(source_type, TimestampType) and isinstance(
            desired_type, DateType
        ):
            return True

        return False

    @staticmethod
    def compare_schema(
        parent_column_name: Optional[str],
        source_schema: DataType,
        desired_schema: DataType,
    ) -> SchemaComparerResult:
        try:
            result: SchemaComparerResult = SchemaComparerResult()
            errors: List[SchemaCompareError] = SchemaComparer.compare_data_type(
                parent_column_name=parent_column_name,
                source_schema=source_schema,
                desired_schema=desired_schema,
                schema_comparison=SchemaComparisonItem(),
                allow_missing_nullable_properties=True,
            )
            # sort the errors by type
            result.errors = (
                [e for e in errors if e.error_type == SchemaCompareErrorType.FATAL]
                + [e for e in errors if e.error_type == SchemaCompareErrorType.ERROR]
                + [e for e in errors if e.error_type == SchemaCompareErrorType.WARNING]
                + [e for e in errors if e.error_type == SchemaCompareErrorType.INFO]
            )

            return result
        except:
            e = sys.exc_info()[0]
            msg = traceback.format_exc()
            raise SparkDataFrameComparerGenericException(e, msg) from e



from os import path, mkdir
from pathlib import Path
from shutil import rmtree

from pyspark.sql import SparkSession, DataFrame

from spark_data_frame_comparer.spark_data_frame_comparer import (
    assert_compare_data_frames,
)
from tests.conftest import clean_spark_session


def test_can_compare_simple(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)
    data_dir: Path = Path(__file__).parent.joinpath("./")
    test_file_path: str = f"{data_dir.joinpath('input/complex.json')}"

    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    mkdir(temp_folder)

    df1: DataFrame = spark_session.read.json(test_file_path)
    df2: DataFrame = spark_session.read.json(test_file_path)

    # Act
    result_path = temp_folder.joinpath("result.json")
    df2.write.json(str(result_path))

    # Assert
    assert_compare_data_frames(
        expected_df=df1,
        result_df=df2,
        expected_path=test_file_path,
        result_path=result_path,
        func_path_modifier=lambda x: x,
    )



from pathlib import Path
from typing import Optional, Union, Callable
from enum import Enum


class ExceptionType(Enum):
    SchemaMismatch = 1
    RowMismatch = 2
    DataMismatch = 3


class SparkDataFrameComparerException(Exception):
    def __init__(
        self,
        exception_type: ExceptionType,
        result: str,
        expected: str,
        message: str,
        additional_info: Optional[str],
        result_path: Optional[Union[Path, str]],
        expected_path: Optional[Union[Path, str]],
        compare_path: Optional[Union[Path, str]],
        func_path_modifier: Optional[
            Callable[[Union[Path, str]], Union[Path, str]]
        ] = None,
    ) -> None:
        self.exception_type: ExceptionType = exception_type
        self.result: str = result
        self.expected: str = expected
        self.message: str = message
        self.result_path: Optional[Union[Path, str]] = result_path
        self.expected_path: Optional[Union[Path, str]] = expected_path
        self.compare_path: Optional[Union[Path, str]] = compare_path
        if func_path_modifier and self.result_path:
            self.result_path = func_path_modifier(self.result_path)
        if func_path_modifier and self.expected_path:
            self.expected_path = func_path_modifier(self.expected_path)
        if func_path_modifier and self.compare_path:
            self.compare_path = func_path_modifier(self.compare_path)
        my_full_message: str = f"""{message}
        {result_path} compared to {expected_path},
        Expected:[{expected}],
        Actual:[{result}]
        Compare: {compare_path or 'No Compare File'}
        Info: {additional_info}
        """
        super().__init__(my_full_message)



from typing import Optional, List


class SchemaComparisonItem:
    def __init__(self) -> None:
        self.source_column: Optional[str] = None
        self.source_column_data_type = None
        self.source_column_is_nullable = None
        self.destination_column: Optional[str] = None
        self.destination_column_data_type = None
        self.destination_column_is_nullable = None

        self.children: Optional[List[SchemaComparisonItem]] = None



from pyspark.sql.types import StructType, StructField, StringType, IntegerType

from spark_data_frame_comparer.schema_comparer import (
    SchemaComparer,
    SchemaComparerResult,
)


def test_schema_comparer() -> None:
    source_schema: StructType = StructType(
        [StructField("name", StringType()), StructField("age", IntegerType())]
    )
    result: SchemaComparerResult = SchemaComparer.compare_schema(
        parent_column_name=None,
        source_schema=source_schema,
        desired_schema=source_schema,
    )

    assert len(result.errors) == 0



from os import path, mkdir
from pathlib import Path
from shutil import rmtree

import pytest
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.functions import lit, col
from pyspark.sql.types import IntegerType

from spark_data_frame_comparer.spark_data_frame_comparer import (
    assert_compare_data_frames,
)
from spark_data_frame_comparer.spark_data_frame_comparer_exception import (
    SparkDataFrameComparerException,
    ExceptionType,
)
from tests.conftest import clean_spark_session


def test_can_compare_simple_fail_with_func_path_modifier(
    spark_session: SparkSession,
) -> None:
    # Arrange
    clean_spark_session(spark_session)
    data_dir: Path = Path(__file__).parent.joinpath("./")
    test_file_path: str = f"{data_dir.joinpath('input/complex.json')}"

    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    mkdir(temp_folder)

    df1: DataFrame = spark_session.read.json(test_file_path)
    df1 = df1.withColumn("simple_field", col("simple_field").cast(IntegerType()))
    df2: DataFrame = spark_session.read.json(test_file_path)
    df2 = df2.withColumn("simple_field", lit(2).cast(IntegerType()))

    # Act
    result_path = temp_folder.joinpath("result.json")
    df2.write.json(str(result_path))

    # Assert
    with pytest.raises(SparkDataFrameComparerException):
        try:
            assert_compare_data_frames(
                expected_df=df1,
                result_df=df2,
                expected_path=test_file_path,
                result_path=result_path,
                temp_folder=temp_folder,
                func_path_modifier=lambda x: Path(
                    str(x).replace(str(data_dir), "/foo/")
                ),
            )
        except SparkDataFrameComparerException as e:
            print(str(e))
            assert e.result_path == Path("/foo/temp/result.json")
            assert e.expected_path == Path("/foo/input/complex.json")
            assert e.compare_path == Path("/foo/temp/compare_complex.json.command")
            assert e.exception_type == ExceptionType.DataMismatch
            raise

    assert path.exists(data_dir.joinpath("temp/compare_complex.json.command"))
    with open(data_dir.joinpath("temp/compare_complex.json.command"), "r") as file:
        line: str = file.read()
        assert (
            line
            == "/usr/local/bin/charm diff "
            + "/foo/temp/result.json"
            + " "
            + "/foo/input/complex.json"
        )



import os
from datetime import datetime
from math import isnan
from pathlib import Path
from typing import List, Any, Tuple, Optional, Union, Callable

from pyspark.sql import DataFrame
from pyspark.sql.types import StructField, ArrayType, StructType, DataType, Row

from spark_data_frame_comparer.list_utils import diff_lists
from spark_data_frame_comparer.spark_data_frame_comparer_exception import (
    SparkDataFrameComparerException,
    ExceptionType,
)


def assert_compare_data_frames(
    expected_df: DataFrame,
    result_df: DataFrame,
    exclude_columns: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    order_by: Optional[List[str]] = None,
    expected_path: Optional[Union[Path, str]] = None,
    result_path: Optional[Union[Path, str]] = None,
    temp_folder: Optional[Union[Path, str]] = None,
    func_path_modifier: Optional[Callable[[Union[Path, str]], Union[Path, str]]] = None,
) -> None:
    """
    Compare two data frames and throws an exception if there is any difference

    :param func_path_modifier:
    :param expected_df:
    :param result_df:
    :param exclude_columns:
    :param include_columns:
    :param order_by:
    :param expected_path:
    :param result_path:
    :param temp_folder:
    :return:
    """
    if exclude_columns:
        result_df = result_df.drop(*exclude_columns)
        expected_df = expected_df.drop(*exclude_columns)
    if order_by:
        expected_df = expected_df.orderBy(*order_by)
        result_df = result_df.orderBy(*order_by)
    if include_columns:
        result_df = result_df.select(*include_columns)
        expected_df = expected_df.select(*include_columns)

    result_df = result_df.select(sorted(result_df.columns))
    expected_df = expected_df.select(sorted(expected_df.columns))

    compare_sh_path: Optional[Path] = None
    if expected_path and result_path and temp_folder:
        expected_file_name: str = os.path.basename(expected_path)
        # create a temp file to launch the diff tool
        # use .command: https://stackoverflow.com/questions/5125907/how-to-run-a-shell-script-in-os-x-by-double-clicking
        compare_sh_path = Path(temp_folder).joinpath(
            f"compare_{expected_file_name}.command"
        )
        with open(compare_sh_path, "w") as compare_sh:
            compare_sh.write(
                f"/usr/local/bin/charm diff "
                f"{func_path_modifier(result_path) if func_path_modifier else result_path} "
                f"{func_path_modifier(expected_path) if func_path_modifier else expected_path}"
            )
            os.fchmod(compare_sh.fileno(), 0o7777)

    if sorted(result_df.columns) != sorted(expected_df.columns):
        print_data_frame_info(expected_df=expected_df, result_df=result_df)
        additional_info = f"Columns not matched:[{diff_lists(expected_df.columns, result_df.columns)}]"
        # remove the compare file since the issue is schema
        if os.path.exists(str(compare_sh_path)):
            os.remove(str(compare_sh_path))
        raise SparkDataFrameComparerException(
            exception_type=ExceptionType.SchemaMismatch,
            result=",".join(result_df.columns),
            expected=",".join(expected_df.columns),
            expected_path=expected_path,
            result_path=result_path,
            compare_path=compare_sh_path,
            message="Columns not matched",
            additional_info=additional_info,
            func_path_modifier=func_path_modifier,
        )

    # compare the types
    result_columns: List[Tuple[str, str]] = list(
        map(lambda t: (t.name, t.dataType.typeName()), result_df.schema)
    )
    expected_columns: List[Tuple[str, str]] = list(
        map(lambda t: (t.name, t.dataType.typeName()), expected_df.schema)
    )
    number_of_mismatched_columns: int = 0
    mismatched_column_errors: List[str] = []
    for i in range(0, len(result_columns)):
        result_column = result_columns[i]
        expected_column = expected_columns[i]
        if result_column != expected_column:
            mismatched_column_errors.append(
                f"column type for {result_column[0]} did not match in {result_path}: "
                f"expected: {expected_column[1]}, actual: {result_column[1]}"
            )

            number_of_mismatched_columns += 1

    if number_of_mismatched_columns > 0:
        print_data_frame_info(expected_df=expected_df, result_df=result_df)
        message = f"{number_of_mismatched_columns} column types did not match"
        # remove the compare file since the issue is schema
        if os.path.exists(str(compare_sh_path)):
            os.remove(str(compare_sh_path))
        raise SparkDataFrameComparerException(
            exception_type=ExceptionType.SchemaMismatch,
            result=",".join(result_df.columns),
            expected=",".join(expected_df.columns),
            expected_path=expected_path,
            result_path=result_path,
            compare_path=compare_sh_path,
            message=message,
            additional_info=",".join(mismatched_column_errors),
            func_path_modifier=func_path_modifier,
        )

    if expected_df.count() != result_df.count():
        print_data_frame_info(expected_df, result_df)
        message = (
            f"Expected {expected_df.count()} rows, actual {result_df.count()} rows"
        )
        raise SparkDataFrameComparerException(
            exception_type=ExceptionType.RowMismatch,
            result=str(result_df.count()),
            expected=str(expected_df.count()),
            expected_path=expected_path,
            result_path=result_path,
            compare_path=compare_sh_path,
            message=message,
            additional_info="",
            func_path_modifier=func_path_modifier,
        )
    error_count: int = 0
    result_rows: List[Row] = result_df.collect()
    expected_rows: List[Row] = expected_df.collect()
    if len(result_rows) != len(expected_rows):
        print_data_frame_info(expected_df, result_df)
        message = f"Expected {len(expected_rows)} rows, actual {len(result_rows)} rows in collect()"
        raise SparkDataFrameComparerException(
            exception_type=ExceptionType.RowMismatch,
            result=str(len(result_rows)),
            expected=str(len(expected_rows)),
            expected_path=expected_path,
            result_path=result_path,
            compare_path=compare_sh_path,
            message=message,
            additional_info="",
            func_path_modifier=func_path_modifier,
        )
    column_schemas: List[StructField] = [t for t in result_df.schema]
    my_errors: List[str] = []
    for row_num in range(0, len(result_rows)):
        if len(expected_rows[row_num]) != len(result_rows[row_num]):
            print_data_frame_info(expected_df, result_df)
            message = (
                f"Expected {len(expected_rows[row_num])} columns, actual {len(result_rows[row_num])} "
                f"columns for row_num: {row_num}"
            )
            raise SparkDataFrameComparerException(
                exception_type=ExceptionType.RowMismatch,
                result=str(len(result_rows[row_num])),
                expected=str(len(expected_rows[row_num])),
                expected_path=expected_path,
                result_path=result_path,
                compare_path=compare_sh_path,
                message=message,
                additional_info="",
                func_path_modifier=func_path_modifier,
            )
        for column_num in range(0, len(result_columns)):
            schema_for_column: StructField = column_schemas[column_num]
            result_value = result_rows[row_num][column_num]
            expected_value = expected_rows[row_num][column_num]
            if (expected_value is None or expected_value == "") and (
                result_value is None or result_value == ""
            ):
                pass
            elif result_value is None or expected_value is None:
                error_count += 1
                my_errors.append(
                    f"row {row_num}: column {column_num} "
                    f"expected: [{expected_value}] actual: [{result_value}]"
                )
            else:
                column_error_count: int
                column_errors: List[str]
                column_error_count, column_errors = check_column_value(
                    column_num=column_num,
                    error_count=error_count,
                    expected_value=expected_value,
                    result_columns=result_columns,
                    result_value=result_value,
                    row_num=row_num,
                    data_type_for_column=schema_for_column.dataType,
                )
                error_count += column_error_count
                my_errors = my_errors + column_errors

    if error_count != 0:
        print_data_frame_info(expected_df=expected_df, result_df=result_df)
        raise SparkDataFrameComparerException(
            exception_type=ExceptionType.DataMismatch,
            result=str(result_df.count()),
            expected=str(expected_df.count()),
            expected_path=expected_path,
            result_path=result_path,
            compare_path=compare_sh_path,
            message=f"{result_path} did not match expected {expected_path}.",
            additional_info=",".join(my_errors),
            func_path_modifier=func_path_modifier,
        )

    # if everything worked then remove the compare file
    if os.path.exists(str(compare_sh_path)):
        os.remove(str(compare_sh_path))


def print_data_frame_info(expected_df: DataFrame, result_df: DataFrame) -> None:
    print("comparing result to expected")
    print("schema for result")
    result_df.printSchema()
    print("schema for expected")
    expected_df.printSchema()
    print("--------- result ---------")
    result_df.show(truncate=False, n=100)
    print("--------- expected ----------")
    expected_df.show(truncate=False, n=100)
    # print("------- difference (result - expected) -------")
    # result_df.subtract(expected_df).show(truncate=False, n=100)
    # print("------- difference (expected - result) -------")
    # expected_df.subtract(result_df).show(truncate=False, n=100)


def check_column_value(
    column_num: int,
    error_count: int,
    expected_value: Any,
    result_columns: List[Tuple[str, str]],
    result_value: Any,
    row_num: int,
    data_type_for_column: DataType,
) -> Tuple[int, List[str]]:
    my_errors: List[str] = []
    if isinstance(data_type_for_column, ArrayType):
        if result_value is None and expected_value is None:
            return error_count, []
        if result_value is None or expected_value is None:
            return error_count + 1, [
                f"Expected array in row:{row_num}, col:{column_num} to be {expected_value} "
                f"but actual is {result_value}"
            ]
        array_item: Row
        for array_item_index in range(0, len(result_value)):
            element_type: DataType = data_type_for_column.elementType
            result_array_item = result_value[array_item_index]
            if len(expected_value) < array_item_index + 1:
                return error_count + 1, [
                    f"Expected row:{row_num}, col:{column_num} has only {len(expected_value)} "
                    f"items but Actual has > {array_item_index + 1}"
                ]
            expected_array_item = expected_value[array_item_index]
            if isinstance(result_array_item, Row):
                column_error_count: int
                column_errors: List[str]
                column_error_count, column_errors = check_column_value(
                    column_num=column_num,
                    error_count=error_count,
                    expected_value=expected_array_item,
                    result_value=result_array_item,
                    result_columns=result_columns,
                    row_num=row_num,
                    data_type_for_column=element_type,
                )
                error_count += column_error_count
                my_errors = my_errors + column_errors

    elif isinstance(data_type_for_column, StructType):
        column_error_count, column_errors = check_struct(
            column_num=column_num,
            error_count=error_count,
            expected_value=expected_value,
            result_value=result_value,
            result_columns=result_columns,
            row_num=row_num,
            data_type_for_column=data_type_for_column,
        )
        error_count += column_error_count
        my_errors = my_errors + column_errors
    else:
        column_error_count, column_errors = check_column_simple_value(
            error_count=error_count,
            expected_value=expected_value,
            result_value=result_value,
            row_num=row_num,
            column_value=result_columns[column_num][0],
        )
        error_count += column_error_count
        my_errors = my_errors + column_errors
    return error_count, my_errors


def check_struct(
    column_num: int,
    error_count: int,
    expected_value: Row,
    result_value: Row,
    result_columns: List[Tuple[str, str]],
    row_num: int,
    data_type_for_column: StructType,
) -> Tuple[int, List[str]]:
    if expected_value is None and result_value is None:
        return error_count, []
    if result_value is None or expected_value is None:
        error_count += 1
        return error_count, [
            f"Expected struct in row:{row_num}, col:{column_num} to be {expected_value} "
            f"but actual is {result_value}"
        ]
    if len(result_value) != len(expected_value):
        return error_count + 1, [
            f"Expected struct in row:{row_num}, col:{column_num} to be {expected_value} "
            f"but actual is {result_value}"
        ]
    my_errors: List[str] = []
    for struct_item_index in range(0, len(result_value)):
        result_struct_item = result_value[struct_item_index]
        expected_struct_item = expected_value[struct_item_index]
        struct_item_type: DataType = data_type_for_column.fields[
            struct_item_index
        ].dataType
        column_error_count: int
        column_errors: List[str]
        column_error_count, column_errors = check_column_value(
            column_num=column_num,
            error_count=error_count,
            expected_value=expected_struct_item,
            result_value=result_struct_item,
            result_columns=result_columns,
            row_num=row_num,
            data_type_for_column=struct_item_type,
        )
        error_count += column_error_count
        my_errors = my_errors + column_errors
    return error_count, my_errors


def check_column_simple_value(
    error_count: int,
    expected_value: Any,
    result_value: Any,
    row_num: int,
    column_value: Any,
) -> Tuple[int, List[str]]:
    result_isnan = (
        isinstance(result_value, float)
        and isinstance(expected_value, float)
        and (isnan(result_value) == isnan(expected_value))
    )
    # if result does not match
    #   and result is not NaN
    #   and both result and expected are not false
    if (
        not compare_scalar(expected_value, result_value)
        and not result_isnan
        and not (not result_value and not expected_value)
    ):
        error_count += 1
        return error_count + 1, [
            f"row {row_num}: column {column_value} "
            + f"expected: [{expected_value}] actual: [{result_value}]"
        ]
    return error_count, []


def compare_scalar(
    expected_value: Union[int, float, bool, str],
    result_value: Union[int, float, bool, str],
) -> bool:
    # for datetime lose the microseconds when comparing
    if isinstance(expected_value, datetime) and isinstance(result_value, datetime):
        return expected_value.replace(microsecond=0) == result_value.replace(
            microsecond=0
        )
    return result_value == expected_value



from pyspark.sql.types import (
    StructType,
    StructField,
    StringType,
    IntegerType,
    FloatType,
)

from spark_data_frame_comparer.schema_comparer import SchemaComparer


def test_schema_comparer_struct_to_struct() -> None:
    print("")
    source_schema: StructType = StructType(
        [
            StructField("name", StructType([StructField("given", FloatType())])),
            StructField("age", IntegerType()),
        ]
    )
    desired_schema: StructType = StructType(
        [
            StructField("name", StructType([StructField("given", StringType())])),
            StructField("age", IntegerType()),
        ]
    )
    SchemaComparer.compare_schema(
        parent_column_name=None,
        source_schema=source_schema,
        desired_schema=desired_schema,
    )



from pyspark.sql.types import ArrayType, FloatType, IntegerType, StructField, StructType

from spark_data_frame_comparer.schema_comparer import (
    SchemaComparer,
    SchemaComparerResult,
)


def test_schema_comparer_array_to_array_allow_missing_nullable_properties_fail() -> None:
    print("")
    source_schema: StructType = StructType(
        [
            StructField(
                "name",
                ArrayType(
                    StructType(
                        [
                            StructField("foo", FloatType()),
                            StructField("bar", FloatType(), nullable=True),
                        ]
                    )
                ),
            ),
            StructField("age", IntegerType()),
        ]
    )
    desired_schema: StructType = StructType(
        [
            StructField(
                "name", ArrayType(StructType([StructField("foo", FloatType())]))
            ),
            StructField("age", IntegerType()),
        ]
    )
    result: SchemaComparerResult = SchemaComparer.compare_schema(
        parent_column_name=None,
        source_schema=source_schema,
        desired_schema=desired_schema,
    )

    assert len(result.errors) == 1
    assert (
        result.errors[0].error
        == "None.name.bar is in source schema but not in destination schema"
    )



from pyspark.sql.types import (
    StructType,
    StructField,
    StringType,
    ArrayType,
    BooleanType,
    TimestampType,
    NullType,
)

from spark_data_frame_comparer.schema_comparer import (
    SchemaComparer,
    SchemaComparerResult,
    SchemaCompareErrorType,
)


def test_schema_comparer_complex_not_match() -> None:
    print("")
    source_schema: ArrayType = ArrayType(
        StructType(
            [
                StructField("id", NullType(), True),
                StructField("type", NullType(), True),
                StructField(
                    "actor",
                    StructType(
                        [
                            StructField("id", NullType(), True),
                            StructField("extension", NullType(), True),
                            StructField("reference", StringType(), False),
                            StructField("type", NullType(), True),
                            StructField("identifier", NullType(), True),
                            StructField("display", NullType(), True),
                        ]
                    ),
                    False,
                ),
                StructField("required", NullType(), True),
                StructField("status", StringType(), False),
                StructField("period", NullType(), True),
                StructField("extension", NullType(), True),
            ]
        ),
        False,
    )

    desired_schema: ArrayType = ArrayType(
        StructType(
            [
                StructField("id", StringType(), True),
                StructField("extension", StringType(), True),
                StructField(
                    "type",
                    ArrayType(
                        StructType(
                            [
                                StructField("id", StringType(), True),
                                StructField("extension", StringType(), True),
                                StructField(
                                    "coding",
                                    ArrayType(
                                        StructType(
                                            [
                                                StructField("id", StringType(), True),
                                                StructField(
                                                    "extension", StringType(), True
                                                ),
                                                StructField(
                                                    "system", StringType(), True
                                                ),
                                                StructField(
                                                    "version", StringType(), True
                                                ),
                                                StructField("code", StringType(), True),
                                                StructField(
                                                    "display", StringType(), True
                                                ),
                                                StructField(
                                                    "userSelected", BooleanType(), True
                                                ),
                                            ]
                                        ),
                                        True,
                                    ),
                                    True,
                                ),
                                StructField("text", StringType(), True),
                            ]
                        ),
                        True,
                    ),
                    True,
                ),
                StructField(
                    "actor",
                    StructType(
                        [
                            StructField("id", StringType(), True),
                            StructField("extension", StringType(), True),
                            StructField("reference", StringType(), True),
                            StructField("type", StringType(), True),
                            StructField(
                                "identifier",
                                StructType(
                                    [
                                        StructField("id", StringType(), True),
                                        StructField("extension", StringType(), True),
                                        StructField("use", StringType(), True),
                                        StructField(
                                            "type",
                                            StructType(
                                                [
                                                    StructField(
                                                        "id", StringType(), True
                                                    ),
                                                    StructField(
                                                        "extension", StringType(), True
                                                    ),
                                                    StructField(
                                                        "coding",
                                                        ArrayType(
                                                            StructType(
                                                                [
                                                                    StructField(
                                                                        "id",
                                                                        StringType(),
                                                                        True,
                                                                    ),
                                                                    StructField(
                                                                        "extension",
                                                                        StringType(),
                                                                        True,
                                                                    ),
                                                                    StructField(
                                                                        "system",
                                                                        StringType(),
                                                                        True,
                                                                    ),
                                                                    StructField(
                                                                        "version",
                                                                        StringType(),
                                                                        True,
                                                                    ),
                                                                    StructField(
                                                                        "code",
                                                                        StringType(),
                                                                        True,
                                                                    ),
                                                                    StructField(
                                                                        "display",
                                                                        StringType(),
                                                                        True,
                                                                    ),
                                                                    StructField(
                                                                        "userSelected",
                                                                        BooleanType(),
                                                                        True,
                                                                    ),
                                                                ]
                                                            ),
                                                            True,
                                                        ),
                                                        True,
                                                    ),
                                                    StructField(
                                                        "text", StringType(), True
                                                    ),
                                                ]
                                            ),
                                            True,
                                        ),
                                        StructField("system", StringType(), True),
                                        StructField("value", StringType(), True),
                                        StructField(
                                            "period",
                                            StructType(
                                                [
                                                    StructField(
                                                        "id", StringType(), True
                                                    ),
                                                    StructField(
                                                        "extension", StringType(), True
                                                    ),
                                                    StructField(
                                                        "start", TimestampType(), True
                                                    ),
                                                    StructField(
                                                        "end", TimestampType(), True
                                                    ),
                                                ]
                                            ),
                                            True,
                                        ),
                                        StructField(
                                            "assigner",
                                            StructType(
                                                [
                                                    StructField(
                                                        "id", StringType(), True
                                                    ),
                                                    StructField(
                                                        "extension", StringType(), True
                                                    ),
                                                    StructField(
                                                        "reference", StringType(), True
                                                    ),
                                                    StructField(
                                                        "type", StringType(), True
                                                    ),
                                                    StructField(
                                                        "identifier",
                                                        StructType(
                                                            [
                                                                StructField(
                                                                    "id",
                                                                    StringType(),
                                                                    True,
                                                                ),
                                                                StructField(
                                                                    "extension",
                                                                    StringType(),
                                                                    True,
                                                                ),
                                                                StructField(
                                                                    "use",
                                                                    StringType(),
                                                                    True,
                                                                ),
                                                                StructField(
                                                                    "type",
                                                                    StructType(
                                                                        [
                                                                            StructField(
                                                                                "id",
                                                                                StringType(),
                                                                                True,
                                                                            )
                                                                        ]
                                                                    ),
                                                                    True,
                                                                ),
                                                                StructField(
                                                                    "system",
                                                                    StringType(),
                                                                    True,
                                                                ),
                                                                StructField(
                                                                    "value",
                                                                    StringType(),
                                                                    True,
                                                                ),
                                                                StructField(
                                                                    "period",
                                                                    StructType(
                                                                        [
                                                                            StructField(
                                                                                "id",
                                                                                StringType(),
                                                                                True,
                                                                            )
                                                                        ]
                                                                    ),
                                                                    True,
                                                                ),
                                                                StructField(
                                                                    "assigner",
                                                                    StructType(
                                                                        [
                                                                            StructField(
                                                                                "id",
                                                                                StringType(),
                                                                                True,
                                                                            )
                                                                        ]
                                                                    ),
                                                                    True,
                                                                ),
                                                            ]
                                                        ),
                                                        True,
                                                    ),
                                                    StructField(
                                                        "display", StringType(), True
                                                    ),
                                                ]
                                            ),
                                            True,
                                        ),
                                    ]
                                ),
                                True,
                            ),
                            StructField("display", StringType(), True),
                        ]
                    ),
                    True,
                ),
                StructField("required", StringType(), True),
                StructField("status", StringType(), True),
                StructField(
                    "period",
                    StructType(
                        [
                            StructField("id", StringType(), True),
                            StructField("extension", StringType(), True),
                            StructField("start", TimestampType(), True),
                            StructField("end", TimestampType(), True),
                        ]
                    ),
                    True,
                ),
            ]
        ),
        True,
    )

    result: SchemaComparerResult = SchemaComparer.compare_schema(
        parent_column_name="identifier",
        source_schema=source_schema,
        desired_schema=desired_schema,
    )

    for error in result.errors:
        print(error)

    assert len(result.errors) == 1, result.errors

    assert result.errors[0].error_type == SchemaCompareErrorType.ERROR
    assert result.errors[0].column == "identifier.extension"



from pyspark.sql.types import (
    StructType,
    StructField,
    StringType,
    IntegerType,
    FloatType,
    ArrayType,
)

from spark_data_frame_comparer.schema_comparer import (
    SchemaComparer,
    SchemaComparerResult,
    SchemaCompareErrorType,
)


def test_schema_comparer_array_to_array_not_match() -> None:
    print("")
    source_schema: StructType = StructType(
        [
            StructField("name", ArrayType(StructField("given2", FloatType()))),
            StructField("age", IntegerType()),
        ]
    )
    desired_schema: StructType = StructType(
        [
            StructField("name", ArrayType(StructField("given", FloatType()))),
            StructField("age", StringType()),
        ]
    )
    result: SchemaComparerResult = SchemaComparer.compare_schema(
        parent_column_name="identifier",
        source_schema=source_schema,
        desired_schema=desired_schema,
    )

    for error in result.errors:
        print(error)

    assert len(result.errors) == 2, result.errors

    assert result.errors[0].error_type == SchemaCompareErrorType.ERROR
    assert result.errors[0].column == "identifier.name"

    assert result.errors[1].error_type == SchemaCompareErrorType.INFO
    assert result.errors[1].column == "identifier.age"
